;;; vc-hooks.el --- resident support for version-control (CS-RCS -- Windows 95/NT)

;; Copyright (C) 1992, 1993, 1994, 1995, 1996, 1997, 1998 Free Software Foundation, Inc.

;; Author: Eric S. Raymond <esr@snark.thyrsus.com>
;; Modified by:
;;   Per Cederqvist <ceder@lysator.liu.se>
;;   Andre Spiegel <spiegel@berlin.informatik.uni-stuttgart.de>
;;   David Matyas <davidm@ComponentSoftware.com>

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; This is the always-loaded portion of VC.
;; It takes care VC-related activities that are done when you visit a file,
;; so that vc.el itself is loaded only when you use a VC command.
;; See the commentary of vc.el.

;;
;;; ComponentSoftware RCS (CS-RCS)
;;
;; CS-RCS is revision control system for Windows 95/NT. Based on GNU RCS, it is fully
;; integrated with the Windows Explorer, providing robust and intuitive version control
;; solution. CS-RCS is free for individuals. For full information and to download your free
;; copy visit: http://www.ComponentSoftware.com/csrcs/
;;
;; For technical support and comments, email: csrcs-support@ComponentSoftware.com
;;

;;
;;; CS-RCS Notes:
;;
;; 1. The CS-RCS implementation uses CS-RCS directly and does NOT use vc.el. 
;;
;; 2. If you wish to exclude RCS support from specific folders, see the comments on
;;    vc-update-state function.
;;    (Note: Use this feature if you encounter performance problems, etc.)
;
;; Known Problems:
;; 1. This scrip sometimes hangs on Windows 95. To proceed, press Ctrl+Alt+Del and terminate
;;    task CSRCSSRV (twice).
;;    Suggested workaround: Keep the CS-RCS Document Explorer open as long as you use Emacs. 
;;

;;
;;; VC.EL Installation:
;;
;; 1. Copy this file into the Emacs "Lisp" directory.
;; 2. Open this file with Emacs.
;; 3. Apply the "Emacs-Lisp | Byte-compile This File" menu command. 
;; 4. Open your "~/.emacs" file (or create a new one if does not exist).
;;    (See below on how to define your home.) 
;; 5. Add the following lines: (adjust vc-path to your machine settings)
;;          (load "vc-hooks")
;;          (setq vc-path "C:/program files/ComponentSoftware/cs-rcs/system")
;; 6. Exit Emacs.
;; 7. Version control context-sensitive commands are now available under the
;;    "Tools | Version Control" menu.
;; 8. Enjoy!
;;

;;
;;; "HOME" directory definition on Windows
;;
;; One way to define your "~" folder is by the following registry key value:
;;       (refer to the NTEmacs FAQ for full details)
;;[HKEY_LOCAL_MACHINE\SOFTWARE\GNU\Emacs]
;;"HOME"="C:\\emacs"
;;

;;; Code:

;; Customization Variables

(defvar f nil
  "False value")

(defvar csrcs "csrcs.exe"
  "CSRCS.EXE path")

(make-variable-buffer-local 'vc-state)
(put 'vc-state 'permanent-local t)

(make-variable-buffer-local 'vc-mode)
(put 'vc-mode 'permanent-local t)

(make-variable-buffer-local 'vc-modified)
(put 'vc-modified 'permanent-local t)

;; We need a notion of per-file properties because the version
;; control state of a file is expensive to derive --- we compute
;; them when the file is initially found, keep them up to date 
;; during any subsequent VC operations, and forget them when
;; the buffer is killed.

(defun vc-simple-command (okstatus command file &rest args)
  ;; Simple version of vc-do-command, for use in vc-hooks only.
  (if (buffer-modified-p) (save-buffer))
  (let 
    (cscommand exec-status)
    (setq cscommand (concat vc-path "\\" command))
    (setq exec-status 
	  (apply 'call-process cscommand nil nil nil 
		 (append args (list file))))
    (cond ((and (> exec-status okstatus) (not (eq -1 okstatus)))
	   (error "RCS command terminated!")))
    exec-status))

(defun vc-simple-command-nowait (command file &rest args)
  ;; Simple version of vc-do-command, for use in vc-hooks only.
  (if (buffer-modified-p) (save-buffer))
  (let 
    (cscommand exec-status)
    (setq cscommand (concat vc-path "\\" command))
    (setq exec-status 
	  (apply 'call-process cscommand nil 0 nil 
		 (append args (list file))))
    exec-status))

(defun vc-update-state ()
  ;; if user has killed the buffer and revisited.
  (setq vc-state (cond
;; Activate next line(s) to exclude RCS support from specific folders.
;; (Note: Use this feature if you encounter performance problems, etc.)
;;		  ((string-match "c:/ExcludeFolder" buffer-file-name) -1)
;;		  ((string-match "c:/temp" buffer-file-name) -1)
		  (t (vc-simple-command -1 csrcs buffer-file-name "UD_State"))
		  ))
	(if ( > vc-state 0 )
	      ( cond ((not vc-make-backup-files)
	      (make-local-variable 'backup-inhibited)
	      (setq backup-inhibited t))))
	(setq vc-mode (cond 
		       ((eq vc-state -1) nil)
		       ((eq vc-state 0) nil)
		       ((eq vc-state 1) ":Archived")
		       ((eq vc-state 2) ":Blocked")
		       ((eq vc-state 3) ":Locked")
		       ((eq vc-state 4) ":Modified")
		       ((eq vc-state 5) ":New")
		       ((eq vc-state 6) ":Updated")
		       ((eq vc-state 7) ":Unlocked-Modified")
		       ((eq vc-state 8) ":Deleted")
		      (t ":RCS Unknown")
	))
;; Uncomment the following line if you are using a shared working folder
;; Force buffer to read-only if file is blocked by another user.
;;  (if (or (eq vc-state 2) (eq vc-state 7)) (setq buffer-read-only t))
  (setq vc-modified (or (eq vc-state 4) (eq vc-state 7)))
  (force-mode-line-update)
)

(defun vc-find-file-hook ()
  ;; Recompute whether file is version controlled,
  ;; if user has killed the buffer and revisited.
  (vc-update-state)
)

;;; install a call to the above as a find-file hook
(add-hook 'find-file-hooks 'vc-find-file-hook)

(defun vc-after-save ()
  ;; Function to be called by basic-save-buffer (in files.el).
  ;; If the file in the current buffer is under version control,
  ;; not locked, and the checkout model for it is `implicit',
  ;; mark it "locked" and redisplay the mode line.
  (vc-update-state)
)

;;; more hooks, this time for file-not-found
(defun vc-file-not-found-hook ()
  "When file is not found, try to check it out from RCS or SCCS.
Returns t if checkout was successful, nil otherwise."
  (interactive)
  (vc-update-state)
  (if vc-mode (vc-co-action 0))
)

(add-hook 'find-file-not-found-hooks 'vc-file-not-found-hook)


;;;###autoload
(defun vc-register (&optional override comment)
  "Register the current file into your version-control system."
  (interactive "P")
  (vc-simple-command 0 csrcs buffer-file-name "add")
  (revert-buffer t t)
  (vc-update-state)
)

;;;###autoload
(defun vc-directory (dirname verbose)
  "Show version-control status of the current directory and subdirectories."
  (interactive "DExplorer (directory): \nP")
  (setq dirname (concat dirname "/*.*"))
  (vc-simple-command 0 csrcs dirname "status" "/s")
)

;;;###autoload
(defun  vc-explorer ()
  "Show version-control status of all files."
  (interactive)
  (vc-simple-command-nowait "documentmanager.exe" "-p")
)

;;;###autoload
(defun vc-next-action (verbose)
  "Do the next logical checkin or checkout operation on the current file.
   CS-RCS - used for check-in only."
  (interactive "P")
  (vc-simple-command 0 csrcs buffer-file-name "checkin")
  (revert-buffer t t)
  (vc-update-state)
)

;;;###autoload
(defun vc-co-action (&optional auto)
  "Do checkout operation on the current file."
;;  (interactive)
  (let 
    (arg1 arg2)
    (if auto (if buffer-read-only (setq arg1 "/l") (setq arg1 "/u")))
;; - better leave confirmation    (if auto (setq arg2 "/q"))
  (vc-simple-command 0 csrcs buffer-file-name "checkout" arg1)
  (revert-buffer t t)
  (vc-update-state)
))

(defun vc-toggle-read-only (&optional verbose)
  "Change read-only status of current buffer, perhaps via version control.
If the buffer is visiting a file registered with version control,
then check the file in or out.  Otherwise, just change the read-only flag
of the buffer.  With prefix argument, ask for version number."
  (interactive "P")
  (if vc-mode
      (vc-co-action t)
    (toggle-read-only)))
(define-key global-map "\C-x\C-q" 'vc-toggle-read-only)

;;;###autoload
(defun vc-revert-buffer ()
  "Revert the current buffer's file back to the latest checked-in version.
This asks for confirmation if the buffer contents are not identical
to that version."
  (interactive)
  (vc-co-action 0)
)

;;;###autoload
(defun vc-diff (historic &optional not-urgent)
  "Display diffs between file versions.
Normally this compares the current file and buffer with the most recent 
checked in version of that file.  This uses no arguments."
  (interactive (list current-prefix-arg t))
  (if historic
      (vc-print-log)
(vc-simple-command 0 csrcs buffer-file-name "status")))

;;;###autoload
(defun vc-print-log ()
  "List the change log of the current buffer in a window."
  (interactive)
  (vc-simple-command 0 csrcs buffer-file-name "history")
)

;;; Now arrange for bindings and autoloading of the main package.
;;; Bindings for this have to go in the global map, as we'll often
;;; want to call them from random buffers.

(setq vc-prefix-map (lookup-key global-map "\C-xv"))
(if (not (keymapp vc-prefix-map))
    (progn
      (setq vc-prefix-map (make-sparse-keymap))
      (define-key global-map "\C-xv" vc-prefix-map)
      (define-key vc-prefix-map "a" 'vc-update-change-log)
      (define-key vc-prefix-map "c" 'vc-cancel-version)
      (define-key vc-prefix-map "d" 'vc-directory)
      (define-key vc-prefix-map "h" 'vc-insert-headers)
      (define-key vc-prefix-map "i" 'vc-register)
      (define-key vc-prefix-map "l" 'vc-print-log)
      (define-key vc-prefix-map "r" 'vc-retrieve-snapshot)
      (define-key vc-prefix-map "s" 'vc-create-snapshot)
      (define-key vc-prefix-map "u" 'vc-revert-buffer)
      (define-key vc-prefix-map "v" 'vc-next-action)
      (define-key vc-prefix-map "=" 'vc-diff)
      (define-key vc-prefix-map "~" 'vc-version-other-window)))

(if (not (boundp 'vc-menu-map))
    ;; Don't do the menu bindings if menu-bar.el wasn't loaded to defvar
    ;; vc-menu-map.
    ()
  ;;(define-key vc-menu-map [show-files]
  ;;  '("Show Files under VC" . (vc-directory t)))
  (define-key vc-menu-map [separator0] '("----"))
  (define-key vc-menu-map [vc-explorer] '("Document Explorer" . vc-explorer))
  (define-key vc-menu-map [vc-directory] '("Show Locked Files" . vc-directory))
  (define-key vc-menu-map [separator1] '("----"))
  (define-key vc-menu-map [vc-rename-file] '("Rename File" . vc-rename-file))
  (define-key vc-menu-map [vc-version-other-window]
    '("Show Other Version" . vc-version-other-window))
  (define-key vc-menu-map [vc-diff] '("Compare with Last Version" . vc-diff))
  (define-key vc-menu-map [vc-update-change-log]
    '("Update ChangeLog" . vc-update-change-log))
  (define-key vc-menu-map [vc-print-log] '("Show History" . vc-print-log))
  (define-key vc-menu-map [separator2] '("----"))
  (define-key vc-menu-map [undo] '("Undo Last Check-In" . vc-cancel-version))
  (define-key vc-menu-map [vc-revert-buffer]
    '("Revert to Last Version" . vc-revert-buffer))
  (define-key vc-menu-map [vc-insert-header]
    '("Insert Header" . vc-insert-headers))
  (define-key vc-menu-map [vc-menu-check-in] '("Check In" . vc-next-action))
  (define-key vc-menu-map [vc-check-out] '("Check Out" . vc-toggle-read-only))
  (define-key vc-menu-map [vc-register] '("Register" . vc-register)))

(put 'vc-rename-file 'menu-enable 'f)
(put 'vc-version-other-window 'menu-enable 'f)
(put 'vc-diff 'menu-enable 'vc-modified)
(put 'vc-update-change-log 'menu-enable 'f) ;; was: '(eq (vc-buffer-backend) 'RCS))
(put 'vc-print-log 'menu-enable 'vc-mode)
(put 'vc-cancel-version 'menu-enable 'f)
(put 'vc-revert-buffer 'menu-enable 'vc-modified)
(put 'vc-insert-headers 'menu-enable 'f)
(put 'vc-next-action 'menu-enable 'vc-modified)
(put 'vc-toggle-read-only 'menu-enable 'vc-mode)
(put 'vc-register 'menu-enable '(and buffer-file-name (eq vc-state 0)))
(put 'vc-directory 'menu-enable 't)

(provide 'vc-hooks)

;;; vc-hooks.el ends here










